#!/usr/bin/env python

"""
    ========
    Genonets
    ========

    This package provides a high level interface for construction and analysis of genotype networks from data.

    Following features are available:

    * Parsing of genotype-phenotype maps from an input file, provided in the genonets input file format
    * Creation of genotype networks from the input data
    * Various analyses on the constructed genotype networks
    * Generation of result files with attributes from genotype network level analyses, as well as genotype level
      analyses
    * Creation of a phenotype network that shows evolvability and accessibility relationships between the genotype sets
    * Generation of GML files corresponding to the created genotype networks and the phenotype network

    Tutorial I: The one-liner
    -------------------------

    The following code snippet shows the simplest possible way of using Genonets::

        from genonets.cmdl_handler import CmdParser  # For parsing command line arguments
        from genonets.genonets_interface import Genonets  # Interface to Genonets API

        Genonets(CmdParser().getArgs(), process=True)

    Yes, that's it. Two import statements, and just one line of code to create, analyze, and save, all the
    genotype networks in the input file, as well the phenotype network. In fact, these are the contents of the
    `genonets_exmpl_minimal.py` sample file included in the package.

    Assuming the you have downloaded the sample code and changed directory to 'sample/', we can run
    `genonets_exmpl_minimal.py` from the command line as follows::

        python genonets_exmpl_minimal.py DNA true data/genonets_sample_input.txt 0.35 results_simple

    The command line arguments specified above are all mandatory positional arguments, i.e., one must specify each
    one of these arguments, and in the correct order. Here's the ordered list of arguments and the corresponding
    descriptions:

    1. **Alphabet type:** The type of alphabet used in the input file. Supported values are:
     * RNA
     * DNA
     * Protein
     * Binary
    2. **Include indels:** Whether or not indels should be considered mutation types.
    3. **Input file name:** Path to, and name of the input file.
    4. **Tau:** The minimum score value to consider when reading genotypes from the input file.
    . **Result directory:** Path to the directory in which the result files should be created.

    Note: These parameters are described in detail in the *Genonets Server Tutorial*.

    In addition to the above listed mandatory arguments, the following optional arguments are also available:

    * **'-rc' or '---use_reverse_complements':** This argument can be specified without a value to enable consideration
      of reverse complements during the construction and analysis of genotype networks. This argument can only be used
      with alphabet type *DNA*.
    * **'-np' or '---num_processes':** The number of processes to use when working in the parallel processing mode. An
      integer value greater than 0 has to be specified as the value to this argument.
    * **'-v' or '---verbose':** This argument can be specified without a value to enable detailed printing of the
      processing steps. Please note that this argument should be used in conjunction with the '-u' option for the
      'python command', e.g.,::

        python -u genonets_exmpl_minimal.py DNA true genonets_sample_input.txt 0.35 results_simple -v


    Tutorial II: Step by step processing
    ------------------------------------

    Instead of using a single call to perform all the processing steps, one can split these steps into multiple
    function calls. Here's a code sample::

        # Parse the command line argument
        args = CmdParser().getArgs()

        # Create the Genonets object. This will load the input file into memory.
        gn = Genonets(args)

        # Use 'gn' to create genotype networks for all genotype sets.
        gn.create()

        # Perform all available analyses on all genotype networks.
        gn.analyze()

        # Write all genotype networks to files in GML format. For a genotype network
        # with two or more components, two files are generated: One corresponds to the
        # entire network with all components, and the other corresponds to the dominant
        # component only.
        gn.save()

        # Save all genotype network level measures to 'Genotype_set_measures.txt'.
        gn.save_network_results()

        # Save all genotype level measures to '<genotypeSetName>_genotype_measures.txt'
        # files. One file per genotype set is generated.
        gn.save_genotype_results()

    Tutorial II: Parallel processing
    --------------------------------

    Parallel processing an be used independently in network creation and network analysis. Here's how parallel
    processing can be enabled::

        # Use 'gn' to create genotype networks for all genotype sets in parallel.
        gn.create(parallel=True)

        # Perform all available analyses on all genotype networks in parallel.
        gn.analyze(parallel= True)

    Tutorial III: Selective processing
    ----------------------------------

    It is also possible to process only a selection of genotype sets from the input data. Also, it is possible to
    perform only a selection of available analysis types. Here's code sample::

        # Parse the command line argument
        args = CmdParser().getArgs()

        # Create the Genonets object. This will load the input file into
        # memory.
        gn = Genonets(args)

        # Use 'gn' to create genotype networks for all genotype sets.
        gn.create()

        # Perform only 'Robustness' and 'Evolvability' analyses on just two of
        # the genotype sets available in the input file, i.e., 'Foxa2' and 'Bbx'.
        gn.analyze(["Foxa2", "Bbx"], analyses=[ac.ROBUSTNESS, ac.EVOLVABILITY])

        # Write the given genotype networks to files in GML format.
        # For a genotype network with two or more components, two files are generated:
        # One corresponds to the entire network with all components, and the other
        # corresponds to the dominant component only.
        gn.save(["Foxa2", "Bbx"])

        # Save genotype network level measures for the given genotype sets to
        # 'Genotype_set_measures.txt'.
        gn.save_network_results(["Foxa2", "Bbx"])

        # Save all genotype level measures for the given genotype sets to
        # 'Foxa2_genotype_measures.txt' and 'Bbx_genotype_measures.txt' files.
        gn.save_genotype_results(["Foxa2", "Bbx"])

    Tutorial III: Customizing results
    ---------------------------------

    This tutorial illustrates the process of customizing the output by adding information to the result file that would
    not be added by Genonets by default.

    The 'Peaks' analysis is used as an example. By default, the 'Peaks' analysis stores results in a dictionary of the
    form::

        {key=peakID : value=[genotypes in the peak]}.

    The sample code that follows customizes this dictionary by adding the score value corresponding to each
    genotype in the list. The resulting dictionary is of the format::

        {key=peakID : value=[(genotype1, score1), ..., (genotypeN, scoreN)]}

    i.e., it is a list of tuples.::

        # Parse the command line argument
        args = CmdParser().getArgs()

        # Create the Genonets object. This will load the input file into
        # memory.
        gn = Genonets(args)

        # Use 'gn' to create genotype networks for all genotype sets.
        gn.create()

        # Perform 'Peaks' analysis
        gn.analyze(analyses=[ac.PEAKS])

        # At this point, the analysis is done. We now need to extract the
        # data we need, i.e., the peaks dictionary.

        # For each genotype set,
        for genotypeSet in gn.genotype_sets():
            # Get the igraph object for the giant
            giant = gn.dominant_network(genotypeSet)

            # Get the dict of peaks {key=peakId : value=[list of sequences in the peak]}
            peaks = giant["Peaks"]

            # Update the dict of peaks, so that instead of just a list of
            # genotypes, we have a list of tuples (sequence, escore).
            newPeaks = {}

            # For each peak,
            for peak in peaks:
                # Initialize the list of tuples
                seqScrTuples = []

                # For each sequence in this peak,
                for sequence in peaks[peak]:
                    # Find the corresponding vertex in the giant
                    try:
                        vertex = giant.vs.find(sequences=sequence)
                    except ValueError:
                        print("Oops! can't find " + sequence + " in giant.")

                    # Get the escore
                    score = giant.vs[vertex.index]["escores"]

                    # Add the tuple to the list of tuples
                    seqScrTuples.append((sequence, score))

                # Add the peak and the corresponding list of tuples to the
                # new peaks dict
                newPeaks[peak] = seqScrTuples

            # Replace the peaks dict in giant with the new dict. This is
            # useful because now if you use the genonets functions below to
            # save the network and results, the updated peaks dict with tuples
            # will be written automatically to file.
            giant["Peaks"] = newPeaks

        # Save networks to file in GML format
        gn.save()

        # Save the results to file from network level analysis
        gn.save_network_results()


    :author:    Fahad Khalid
    :license:   The MIT License (MIT)

                Copyright (c) 2016 Fahad Khalid and Joshua L. Payne

                Permission is hereby granted, free of charge, to any person obtaining a copy of this software and
                associated documentation files (the "Software"), to deal in the Software without restriction, including
                without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
                copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to
                the following conditions:

                The above copyright notice and this permission notice shall be included in all copies or substantial
                portions of the Software.

                THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
                LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
                NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
                WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
                SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""



"""
    genonets_filters
    ~~~~~~~~~~~~~~~~

    Contains filters used throughout the package.

    :author: Fahad Khalid
    :license: MIT, see LICENSE for more details.
"""


class WriterFilter:
    def __init__(self):
        pass

    # Ordered list of genotype sets, initialized by the Genonets object
    ORDERED_GENOTYPE_SETS = None

    # Network level attribute to order map
    ordered_net_attrib = {
        "Robustness": 1,
        "Evolvability": 2,
        "Evolvability_targets": 3,
        "Accessibility": 4,
        "Neighbor_abundance": 5,
        "Diversity_index": 6,
        "Number_of_genotype_networks": 7,
        "Genotype_network_sizes": 8,
        "Size_of_dominant_genotype_network": 9,
        "Proportional_size_of_dominant_genotype_network": 10,
        "Diameter": 11,
        "Edge_density": 12,
        "Assortativity": 13,
        "Average_clustering_coefficient_of_dominant_genotype_network": 14,
        "Number_of_peaks": 15,
        "Peaks": 16,
        "Number_of_squares": 17,
        "Magnitude_epistasis": 18,
        "Simple_sign_epistasis": 19,
        "Reciprocal_sign_epistasis": 20,
        "Ratio_of_overlapping_genotype_sets": 21,
        "Overlapping_genotype_sets": 22,
        "Epistasis_squares": 23
    }

    # Genotype level attribute to order map
    ordered_seq_attrib = {
        "Robustness": 1,
        "Evolvability": 2,
        "Evolvability_targets": 3,
        "Evolves_to_genotypes_in": 4,
        "Overlaps_with_genotypes_in": 5,
        "Coreness": 6,
        "Clustering_coefficient": 7,
        "Distance from Summit": 8,
        "Accessible_paths_through": 9
    }

    @staticmethod
    def gmlAttribsToIgnore(level):
        if level == "network":
            attrs = [
                "Evolvability_targets",
                "SqrEpi_list",
                "diameterPath_list",
                "Squares_list"
            ]
        elif level == "vertex":
            attrs = [
                "name",
                "label",
                "pathsToSummit",
                "VtxToSqrs"
            ]

        return attrs

    @staticmethod
    def netAttribsToIgnore():
        return [
            "name",
            "SqrEpi_list",
            "diameterPath_list",
            "Squares_list",
            "Summit"
        ]

    @staticmethod
    def seqAttribsToIgnore():
        return [
            "name",
            "sequences",
            "label",
            "escores",
            "pathsToSummit",
            "VtxToSqrs"
        ]

    @staticmethod
    def net_attribute_to_order(attribute):
        try:
            return WriterFilter.ordered_net_attrib[attribute]
        except KeyError:
            # Any custom attribute added by the user will be assigned
            # this arbitrary high value, so that the user-defined
            # attributes are placed at the very end.
            return 1000

    @staticmethod
    def seq_attribute_to_order(attribute):
        try:
            return WriterFilter.ordered_seq_attrib[attribute]
        except KeyError:
            # Any custom attribute added by the user will be assigned
            # this arbitrary high value, so that the user-defined
            # attributes are placed at the very end.
            return 1000

    @staticmethod
    def genotype_set_to_order(genotype_set):
        try:
            return WriterFilter.ORDERED_GENOTYPE_SETS.index(genotype_set)
        except ValueError:
            # If the requested genotype set was not found in the list, just
            # return an arbitrarily high value.
            return 9999999



"""
    genonets_reader
    ~~~~~~~~~~~~~~~

    Parser for the genonets input file format.

    :author: Fahad Khalid
    :license: MIT, see LICENSE for more details.
"""

import csv

from genonets_exceptions import GenonetsError
from genonets_constants import ErrorCodes
from genonets_constants import SupportedAlphabet


class InReader:
    def __init__(self):
        pass

    # Required input file column headers
    COL_HEADERS = ["Genotypeset", "Delta", "Genotype", "Score"]

    # Uses the given DictReader to construct a dictionary from file with,
    # key=Genotypeset, value=dict{key=genotype, value=score}
    # Returns the dictionary
    # TODO: This function can be extend to also return a reverse dictionary
    # with key=score, value=sequence ...
    @staticmethod
    def build_data_dicts(in_file_path, tau, alphabet_type):
        # Data structures to be returned
        data_dict = {}
        delta_dict = {}
        genotypes = []      # List of unique genotypes across all genotype sets
        genotype_sets = []  # List of genotype sets in the order in which they are read from file

        # Genotype length to be determined
        genotype_length = 0

        # Get handle to the input file and a DictReader for the file
        reader, in_file = InReader.dict_reader_for_file(in_file_path)

        # Check if all the required column headers are available in the file
        if not InReader.req_hdrs_are_present(reader.fieldnames):
            in_file.close()

            print("Error: " +
                  ErrorCodes.getErrDescription(ErrorCodes.INCONSISTENT_HEADER))

            raise GenonetsError(ErrorCodes.INCONSISTENT_HEADER)

        # For each data row in the file,
        for row in reader:
            # Check for missing values in this row
            if any(row[col] in (None, "") for col in row.keys()):
                in_file.close()

                line_number = str(int(reader.line_num))

                print("Error: " +
                      ErrorCodes.getErrDescription(ErrorCodes.MISSING_VALUE) +
                      ": Line No. " + line_number)

                raise GenonetsError(ErrorCodes.MISSING_VALUE,
                                    "Line No. " + line_number)

            # Get the fitness score
            try:
                score = float(row["Score"])
            except:
                in_file.close()

                line_number = str(int(reader.line_num))

                print("Error: " +
                      ErrorCodes.getErrDescription(ErrorCodes.BAD_SCORE_FORMAT) +
                      ": Line No. " + line_number)

                raise GenonetsError(ErrorCodes.BAD_SCORE_FORMAT,
                                    "Line No. " + line_number)

            # If the score for this genotype is greater than or equal to
            # the given threshold,
            if score >= tau:
                # If the current genotype set has not already been added,
                if row["Genotypeset"] not in data_dict:
                    # Initialize dict for the genotype set
                    data_dict[row["Genotypeset"]] = {}

                    # Add the name of this genotype set to the ordered list of
                    # genotype set names
                    genotype_sets.append(row["Genotypeset"])

                    # Get the delta value
                    try:
                        delta = float(row["Delta"])
                    except:
                        in_file.close()

                        line_number = str(int(reader.line_num))

                        print("Error: " +
                              ErrorCodes.getErrDescription(ErrorCodes.BAD_DELTA_FORMAT) +
                              ": Line No. " + line_number)

                        raise GenonetsError(
                            ErrorCodes.BAD_DELTA_FORMAT,
                            "Line No. " + line_number)

                    delta_dict[row["Genotypeset"]] = delta

                # Get the genotype sequence
                genotype = row["Genotype"]

                # If genotype length has not been initialized yet, i.e., this is
                # the first row,
                if genotype_length == 0:
                    # Set length of the current genotype as the genotype length
                    # for the entire dataset
                    genotype_length = len(genotype)

                try:
                    InReader.verify_genotype(genotype, genotype_length, alphabet_type,
                                             str(int(reader.line_num)))
                except Exception as e:
                    in_file.close()
                    raise e

                # Add genotype as key and score as value to the current
                # genotype set
                data_dict[row["Genotypeset"]][genotype] = score

                # If the genotype has not already been read for any other
                # genotype set,
                if genotype not in genotypes:
                    # Add it to the list of unique sequences found in the
                    # input file
                    genotypes.append(genotype)

        in_file.close()

        # If no genotypes were found with score >= tau,
        if not data_dict:
            print("Error: " +
                  ErrorCodes.getErrDescription(ErrorCodes.NO_USABLE_SCORES) +
                  ": Tau=" + str(tau))

            raise GenonetsError(
                ErrorCodes.NO_USABLE_SCORES,
                "Tau=" + str(tau))

        # Dictionary: Key=Sequence, Value=[Genotype sets]. Reverse dictionary
        # that is used in functions like evolvability.
        genotype_to_set_dict = InReader.build_genotype_to_set_dict(genotypes, data_dict)

        return data_dict, delta_dict, genotype_to_set_dict, genotype_length, genotype_sets

    # Builds Dictionary: Key=Sequence, Value=[Genotype sets in which this sequence
    # exists]
    @staticmethod
    def build_genotype_to_set_dict(genotypes, data_dict):
        genotype_to_set_dict = {}

        # For each sequence in the given list of genotypes,
        for seq in genotypes:
            # Initialize the list genotype sets
            genotype_to_set_dict[seq] = []

            # For each genotype set in the given data dictionary,
            for rep in data_dict.keys():
                # If the sequence exists in the genotype set,
                if seq in data_dict[rep]:
                    # Append the genotype set name to the list of values
                    # for this sequence
                    genotype_to_set_dict[seq].append(rep)

        return genotype_to_set_dict

    # If any of the required column headers are not in the file, returns
    # False. Otherwise return True.
    @staticmethod
    def req_hdrs_are_present(col_names):
        cols_in_file = [col for col in col_names]

        if any(header not in cols_in_file for header in InReader.COL_HEADERS):
            return False

        return True

    @staticmethod
    def verify_genotype(genotype, genotype_length, alphabet_type, line_number):
        # Verify the length
        if len(genotype) != genotype_length:
            print("Error: " +
                  ErrorCodes.getErrDescription(ErrorCodes.INCONSISTENT_SEQ_LEN) +
                  ": Line No. " + line_number)

            raise GenonetsError(
                ErrorCodes.INCONSISTENT_SEQ_LEN,
                "Line No. " + line_number)

        # Get the alphabet corresponding to the type received as
        # argument
        alphabet = SupportedAlphabet.getAlphabet(alphabet_type)

        # Verify alphabet
        if any(letter not in alphabet for letter in genotype):
            print("Error: " +
                  ErrorCodes.getErrDescription(ErrorCodes.ALPHABET_TYPE_MISMATCH) +
                  ": Line No. " + line_number)

            raise GenonetsError(
                ErrorCodes.ALPHABET_TYPE_MISMATCH,
                "Line No. " + line_number)

    # Open the given CSV file, and return a handle to the dict reader
    # for the file, as well as the file handle.
    @staticmethod
    def dict_reader_for_file(file_name):
        # Open file
        try:
            data_file = open(file_name, 'rU')
        except Exception as e:
            print("Error: " +
                  ErrorCodes.getErrDescription(ErrorCodes.UNKNOWN_PARSING_ERROR))

            raise GenonetsError(ErrorCodes.UNKNOWN_PARSING_ERROR)

        # Read the file into a dictionary
        reader = csv.DictReader(data_file, delimiter="\t")

        return reader, data_file



"""
    robustness_functions
    ~~~~~~~~~~~~~~~~~~~~

    Contains functions used for genotype and phenotype robustness computations.

    :author: Fahad Khalid
    :license: MIT, see LICENSE for more details.
"""

import numpy as np


class RobustnessAnalyzer:
    # Constructor
    def __init__(self, network, netBuilder, is_double_stranded):
        # Reference to the network on which to perform this
        # analysis
        self.network = network

        # Get a reference to the NetworkUtils object
        self.netBuilder = netBuilder

        # Get a reference to the BitSeqManipulator in use
        self.bitManip = netBuilder.bitManip

        # Flag indicating whether the genotypes correspond to double
        # stranded molecules
        self.is_double_stranded = is_double_stranded

        # Reference to the list of tuples, (seq, robustness)
        # populated when required
        self.robustnessVals = None

    # Returns the arithmetic mean of the all values in the
    # robustness list.
    def getAvgRobustness(self):
        # Calculate the mean of all robustness values for the network
        return np.mean(self.getRobustnessAll())

    # Returnes a list of tuples of the form (sequence, robustness) that
    # consist of robustness values for all sequences represented by
    # vertices of the given network.
    def getRobustnessAll(self, recompute=False):
        # If either the robustness values have not been computed already,
        # or the caller has explicitly requested re-computing,
        if not self.robustnessVals or recompute:
            # Get robustness values for all sequences in the network
            self.robustnessVals = [
                self.getGenotypeRobustness(seq)
                for seq in self.network.vs["sequences"]
            ]

        return self.robustnessVals

    # Calculates sequence robustness, which is defined as Genotype
    # robustness in,
    # Ref: "Robustness and evolvability: a paradox resolved". Andreas
    # Wagner, Proc. R. Soc. B 2008 275 91-100;
    # DOI: 10.1098/rspb.2007.1137. Published 7 January 2008
    def getGenotypeRobustness(self, sequence):
        # Get the vertex that corresponds to this sequence
        vertex = self.netBuilder.getVertex(sequence, self.network)

        # Get vertex degree
        degree = self.network.degree(vertex)

        # Get a list of all possible 1-neighbors for this sequence
        allNeighbors = self.bitManip.generateNeighbors(
            self.bitManip.seqToBits(sequence))

        # If reverse complements should be considered,
        if self.is_double_stranded:
            # Reverse complement of the focal genotype
            rc = self.bitManip.getReverseComplement(
                self.bitManip.seqToBits(sequence))

            # If the reverse complement is also a 1-mutant,
            if rc in allNeighbors:
                # Do not consider it as one of all possible 1-mutants
                allNeighbors.remove(rc)

            # Make sure a genotype and its reverse complement are not both
            # considered separate neighbors
            for neighbor in list(allNeighbors):
                # Reverse complement of the neighbor
                rc_n = self.bitManip.getReverseComplement(neighbor)

                # If the reverse complement is not the same as the neighbor,
                # and the reverse complement is already in the list of
                # external neighbors, remove the neighbor itself from the
                # list.
                if rc_n != neighbor and rc_n in allNeighbors:
                    allNeighbors.remove(neighbor)

        # Count the number of all possible 1-neighbors of this sequence
        numNeighbors = len(allNeighbors)

        try:
            return float(degree) / float(numNeighbors)
        except ZeroDivisionError:
            return 0



#!/usr/bin/env python

"""
    genonets_exmpl_minimal
    ~~~~~~~~~~~~~~~~~~~~~~

    Demonstrates the minimal code required to create genotype networks, perform analyses,
    and write results to files using Genonets with default settings.

    Use the following command to run the script:
    'python genonets_exmpl_minimal.py DNA true data/genonets_sample_input.txt 0.35 results_minimal'

    Output files will be generated in 'results_minimal/'

    :author: Fahad Khalid
    :license: MIT, see LICENSE for more details.
"""

from genonets.cmdl_handler import CmdParser  # For parsing command line arguments
from genonets.genonets_interface import Genonets  # Interface to Genonets API


if __name__ == "__main__":
    # This single line of code,
    #   1. Parses the command line arguments.
    #   2. Loads the input file into memory.
    #   3. Creates genotype networks corresponding to all genotype sets
    #      available in the input data set.
    #   4. Performs all available analyses on all genotype networks.
    #   5. Writes all genotype networks to files in GML format.
    #   6. Writes all genotype network level results to file.
    #   7. Writes all genotype level results to files.
    #   8. Creates the phenotype network.
    #   9. Writes the phenotype network to file in GML format.
    Genonets(CmdParser().getArgs(), process=True)

    # Print message to indicate processing is done.
    print("\nDone.\n")



"""
    Public interface to Genonets functions.

    :author: Fahad Khalid
    :license: MIT, see LICENSE for more details.
"""

import sys
import copy
from multiprocessing import Process, Queue

from cmdl_handler import CmdArgs
from genonets_writer import Writer
from genonets_reader import InReader
from graph_utils import NetworkBuilder
from seq_bit_impl import BitManipFactory
from genonets_filters import WriterFilter
from analysis_handler import AnalysisHandler
from genonets_constants import ErrorCodes
from genonets_exceptions import GenonetsError
from genonets_constants import GenonetsConstants as Gc
from genonets_constants import AnalysisConstants as Ac


class Genonets:
    """
    Encapsulates the Genonets public API.
    """

    ALL = 0

    def __init__(self, args, process=False, parallel=False):
        """
        Initiate parsing of the input file, and load the parsed data into a `Genonets` object.

        A simple way to create a `Genonets` object is the following::

            gn = Genonets(CmdParser(args).getArgs())

        where, `CmdParser` can be imported as follows::

            from genonets.cmdl_handler import CmdParser

        The `args` variable is a list of command line arguments. Please see the `genonets` package
        documentation for the list and descriptions of all available command line arguments.

        :param args: A populated `CmdArgs` object.
        :param process: If 'True', in addition to creating the object, initiates complete processing, i,e., creates
                        genotype networks for all genotype sets in the input data, performs all available analyses on
                        all genotype networks, and generates all result files.
        :param parallel: Flag to indicate whether or not parallel processing should be used. This parameter is only
                         useful with 'process=True'.
        """

        # Handle program arguments
        self.cmdArgs = CmdArgs(args)

        # Read file and load input data into memory
        self.inDataDict, self.deltaDict, self.seqToRepDict, self.seqLength, self.ordered_genotype_sets = \
            self._build_data_dicts(self.cmdArgs.inFilePath)

        # Pass the ordered list of genotype sets to the WriterFilter
        WriterFilter.ORDERED_GENOTYPE_SETS = self.ordered_genotype_sets

        # Get the bit-sequence manipulator object corresponding to the
        # given molecule type.
        self.bitManip = self._bit_manipulator()

        # Get the NetworkBuilder object
        self.netBuilder = NetworkBuilder(self.bitManip, self.cmdArgs.use_reverse_complements)

        # Dictionary: Key=Repertoire, Value=Network. Created when required.
        self.repToNetDict = {}

        # Dictionary: Key=Repertoire, Value=Giant. Created when required.
        # If there is only one component, giant=network
        self.repToGiantDict = {}

        # Phenotype network
        self.pheno_net = None

        # Reference to the analyzer object
        self.analyzer = None

        # Set the VERBOSE flag
        self.VERBOSE = True if self.cmdArgs.verbose else False

        # If the user has requested complete processing with default settings,
        if process:
            # Perform all processing steps
            self._process_all(parallel)

    def create(self, genotype_sets=Gc.ALL, parallel=False):
        """
        Create genotype networks for the given list of genotype set names.

        :param genotype_sets: List of names of the genotype sets for which the genotype
                             networks should be created. If a value is not explicitly
                             specified for this parameter, genotype networks are
                             constructed for all genotype sets available in the parsed
                             data.
        :param parallel: Flag to indicate whether or not parallel processing should
                         be used.
        :return: No return value
        """

        if self.VERBOSE:
            print("Creating genotype networks:")

        # If a single string is received, convert it into an iterable
        genotype_sets = [genotype_sets] if type(genotype_sets) == str else genotype_sets

        # If all genotype_sets should be considered,
        if genotype_sets == Gc.ALL:
            # Get a list of all genotype_sets
            genotype_sets = self.genotype_sets()

        # If multiprocessing should be used,
        if parallel:
            self._create_networks_parallel(genotype_sets)
        else:
            self._create_networks(genotype_sets)

        if self.VERBOSE:
            sys.stdout.write("Done.\n")

    def analyze(self, genotype_sets=Gc.ALL, analyses=Gc.ALL, parallel=False):
        """
        Performs all analyses provided in the list of analysis types, on the given genotype sets.

        This method can only be used if `create` has already been called on the same `Genonets`
        object.

        :param genotype_sets: List of names of the genotype sets for which the genotype
                            networks should be created. If a value is not explicitly
                            specified for this parameter, genotype networks are
                            constructed for all genotype sets available in the parsed
                            data.
        :param analyses: List of analysis type constants. These constants are defined in the class
                         `genonets.genonets_constants.AnalysisConstants`. If the value for this
                         parameter is not explicitly set, all available analyses are performed.
        :param parallel: Flag to indicate whether or not parallel processing should
                         be used.
        :return: No return value.
        """

        if self.VERBOSE:
            sys.stdout.write("\nPerforming analyses:")

        # If all genotype_sets should be considered,
        if genotype_sets == Gc.ALL:
            # Get a list of all genotype_sets
            genotype_sets = self.genotype_sets()

        # If a single string is received, convert it into an iterable
        genotype_sets = [genotype_sets] if type(genotype_sets) == str else genotype_sets

        # If overlap in one of the requested analyses, there need to be at
        # at least two genotype_sets in the dataset
        if analyses == Gc.ALL or Ac.OVERLAP in analyses:
            if len(genotype_sets) < 2:
                print("Error: " +
                      ErrorCodes.getErrDescription(ErrorCodes.NOT_ENOUGH_REPS_OLAP) +
                      ": Tau=" + str(self.cmdArgs.tau))

                raise GenonetsError(
                    ErrorCodes.NOT_ENOUGH_REPS_OLAP,
                    "Tau=" + str(self.cmdArgs.tau))

        # If multiprocessing should be used,
        if parallel:
            # Perform all analyses in parallel; overlap will be ignored.
            self._analyze_networks_parallel(genotype_sets, analyses)

            if analyses == Gc.ALL or Ac.OVERLAP in analyses:
                # Reset analysis handler to make sure it references
                # the updated dicts
                del self.analyzer
                self.analyzer = AnalysisHandler(self)

                # Use serial processing to perform overlap analysis
                self._analyze_networks(genotype_sets, [Ac.OVERLAP])
        else:
            # Perform all analyses using serial processing
            self._analyze_networks(genotype_sets, analyses)

    def phenotype_network(self, collection_name="phenotype_network", genotype_sets=Gc.ALL):
        """
        Create the phenotype network from the given list of genotype sets.

        :param collection_name: The name to be assigned to the phenotype network.
        :param genotype_sets: List of names of the genotype sets for which the phenotype
                              network should be created. If a value is not explicitly
                              specified for this parameter, all available genotype sets
                              are considered.
        :return: `igraph.Graph` object representing the phenotype network.
        """

        # If a single string is received, convert it into an iterable
        genotype_sets = [genotype_sets] if type(genotype_sets) == str else genotype_sets

        # If all genotype_sets should be considered,
        if genotype_sets == Gc.ALL:
            # Get a list of all genotype_sets
            genotype_sets = self.genotype_sets()

        # Create a list of giants for the given genotype_sets
        giants = [self.repToGiantDict[repertoire] for repertoire in genotype_sets]

        # Create the phenotype network, and get the igraph object
        self.pheno_net = self.netBuilder.createEvoNet(collection_name, giants)

        return self.pheno_net

    def genotype_sets(self):
        """
        Get a list of names of all genotype sets for which genotype networks have been created.

        :return: List of names of genotype sets.
        """
        repertoires = self.inDataDict.keys()
        repertoires = [repertoires] if type(repertoires) == str else repertoires

        return repertoires

    def genotype_network(self, genotype_set):
        """target="_blank"
        Get the `igraph` object for the network corresponding to the given genotype set name.

        The `igraph` object in this case refers to the entire network, i.e., all connected
        components.

        Note: This method can only be used if the genotype network corresponding to the requested
        genotype set name has already been created.

        :param genotype_set: Name of the genotype set for which the genotype network is
                             requested.
        :return: Object of type `igraph.Graph`.
        """

        try:
            return self.repToNetDict[genotype_set]
        except KeyError:
            return None

    def dominant_network(self, genotype_set):
        """
        Get the `igraph` object for the *dominant* network corresponding to the given genotype set name.

        The dominant network refers to the giant component in the network.

        Note: This method can only be used if the genotype network corresponding to the requested
        genotype set name has already been created.

        :param genotype_set: Name of the genotype set for which the genotype network is
                             requested.
        :return: Object of type `igraph.Graph`.
        """

        try:
            return self.repToGiantDict[genotype_set]
        except KeyError:
            return None

    def save(self, genotype_sets=Gc.ALL):
        """
        Write the genotype networks corresponding to the given genotype sets to file.

        The networks are saved in GML format. For networks with more than one
        components, separate files are generated for the entire network and the
        dominant network.

        Note: This method can be used only after `analyze()` has been called on the
        given genotype sets.

        :param genotype_sets: List of names of genotype sets for which the genotype
                              should be written to file. If a value is not explicitly specified
                              for this parameter, result files are written for all
                              genotype sets.
        :return: No return value.
        """

        if self.VERBOSE:
            sys.stdout.write("\nWriting GML files for genotype networks ... ")

        # If a single string is received, convert it into an iterable
        genotype_sets = [genotype_sets] if type(genotype_sets) == str else genotype_sets

        Writer.writeNetsToFile(self.repToNetDict, self.repToGiantDict,
                               self.netBuilder, self.cmdArgs.outPath,
                               WriterFilter.gmlAttribsToIgnore, genotype_sets)

        if self.VERBOSE:
            sys.stdout.write("Done.\n")

    def save_network_results(self, genotype_sets=Gc.ALL):
        """
        Write the genotype set level results to file.

        A file named 'Genotype_set_measures.txt' is generated in the output directory
        specified at the time of the `Genonets` object creation.

        Note: This method can be used only after `analyze()` has been called on the
        given genotype sets.

        :param genotype_sets: List of names of genotype sets for which to generate the
                              result files. If a value is not explicitly specified
                              for this parameter, result files are written for all
                              genotype sets.
        :return: No return value.
        """

        if self.VERBOSE:
            sys.stdout.write("\nWriting genotype set level results ... ")

        # If a single string is received, convert it into an iterable
        genotype_sets = [genotype_sets] if type(genotype_sets) == str else genotype_sets

        Writer.writeNetAttribs(self.repToNetDict, self.repToGiantDict,
                               self.netBuilder, self.cmdArgs.outPath,
                               WriterFilter.netAttribsToIgnore,
                               WriterFilter.net_attribute_to_order,
                               WriterFilter.genotype_set_to_order,
                               genotype_sets)

        if self.VERBOSE:
            sys.stdout.write("Done.\n")

    def save_genotype_results(self, genotype_sets=Gc.ALL):
        """
        Write the genotype level results to files.

        A results file is generated for each genotype set.

        Note: This method can be used only after `analyze()` has been called on the
        given genotype sets.

        :param genotype_sets: List of names of genotype sets for which to generate the
                              result files. If a value is not explicitly specified
                              for this parameter, result files are written for all
                              genotype sets.
        :return: No return value.
        """

        if self.VERBOSE:
            sys.stdout.write("\nWriting genotype level results ... ")

        # If a single string is received, convert it into an iterable
        genotype_sets = [genotype_sets] if type(genotype_sets) == str else genotype_sets

        Writer.writeSeqAttribs(self.repToGiantDict,
                               self.cmdArgs.outPath,
                               WriterFilter.seqAttribsToIgnore,
                               WriterFilter.seq_attribute_to_order,
                               genotype_sets)

        if self.VERBOSE:
            sys.stdout.write("Done.\n")

    def save_phenotype_network(self):
        """
        Write the phenotype network to file in GML format.

        Note: This method can only be used after the phenotype network has been created.

        :return: No return value.
        """

        if self.VERBOSE:
            sys.stdout.write("\nWriting GML file for phenotype network ... ")

        Writer.writeNetToFile(self.pheno_net, self.cmdArgs.outPath,
                              WriterFilter.gmlAttribsToIgnore)

        if self.VERBOSE:
            sys.stdout.write("Done.\n")

    # Plots the given network.
    def plot(self, network, layout="auto"):
        self.netBuilder.plotNetwork(network, layout, self.cmdArgs.outPath)

    # ----------------------------------------------------------------------
    #   Private methods, i.e., those that are not supposed to be part of
    #   public interface
    # ----------------------------------------------------------------------

    def _build_data_dicts(self, inFilePath):
        return InReader.build_data_dicts(inFilePath, self.cmdArgs.tau,
                                         self.cmdArgs.moleculeType)

    def _bit_manipulator(self):
        return BitManipFactory.getBitSeqManip(self.cmdArgs.moleculeType,
                                              self.seqLength,
                                              self.cmdArgs.useIndels,
                                              self.cmdArgs.use_reverse_complements)

    def _bitseqs_and_scores(self, repertoire):
        # Get the list of sequences for the given repertoire
        sequences = self.inDataDict[repertoire].keys()

        # Get the list of scores for the given repertoire
        scores = [self.inDataDict[repertoire][seq] for seq in sequences]

        return sequences, scores

    # Carries out all default processing steps with default settings.
    def _process_all(self, parallel):
        self.create(parallel=parallel)
        self.analyze(parallel=parallel)
        self.save()
        self.save_network_results()
        self.save_genotype_results()
        self.phenotype_network()
        self.save_phenotype_network()

    # Create genotype networks for the given, or all repertoires
    def _create_networks(self, repertoires):
        # For each repertoire,
        for repertoire in repertoires:
            if self.VERBOSE:
                sys.stdout.write(repertoire + " ... ")

            # Get the sequences and scores
            seqs, scores = self._bitseqs_and_scores(repertoire)

            # Create the genotype network and store it in a
            # dictionary: Key=Repertoire, Value=Network
            self.repToNetDict[repertoire] = \
                self.netBuilder.createGenoNet(repertoire, seqs, scores)

            # Get the number of components in the network
            numComponents = len(self.netBuilder.getComponents(
                self.repToNetDict[repertoire]))

            # If there are more than one components,
            if numComponents > 1:
                # Get the giant component
                giant = self.netBuilder.getGiantComponent(self.repToNetDict[repertoire])

                # Set 'name' attribute for giant
                giant["name"] = repertoire + "_dominant"

                # Reference to the giant component
                self.repToGiantDict[repertoire] = giant
            else:
                # The network and giant component are the same
                self.repToGiantDict[repertoire] = self.repToNetDict[repertoire]

    # Use multiprocessing to create genotype networks
    # for the given, or all repertoires
    def _create_networks_parallel(self, repertoires):
        # Instantiate a concurrent queue for results
        resultsQueue = Queue()

        # Compute indices to be used in the loop
        indices = Genonets._process_blocks(len(repertoires), self.cmdArgs.num_procs)

        for i in indices:
            # Create separate processes for each repertoire in the current block
            processes = [
                Process(target=Genonets._create_gn,
                        args=(self.inDataDict[repertoires[j]],
                              self.cmdArgs,
                              self.seqLength,
                              resultsQueue,
                              repertoires[j])
                        )
                for j in range(i - 1, Genonets._len_finished_reps(
                    i, len(repertoires), self.cmdArgs.num_procs))
            ]

            # Start the processes
            for p in processes:
                p.start()

            # Spin lock
            # FIXME: The condition in the loop can result in an infinite
            # iteration if one of the processes does not put results
            # in the queue. This condition should be replaced
            # with one that is reliable ...
            while len(self.repToNetDict) != Genonets._len_finished_reps(
                    i, len(repertoires), self.cmdArgs.num_procs):
                result = resultsQueue.get()

                if self.VERBOSE:
                    sys.stdout.write(result[0] + " ... ")

                self.repToNetDict[result[0]] = result[1][0]
                self.repToGiantDict[result[0]] = result[1][1]

    @staticmethod
    def _create_gn(seqScrDict, args, seqLength, resultsQueue, repertoire):
        # Get a reference to the bit manipulator
        bitManip = BitManipFactory.getBitSeqManip(args.moleculeType,
                                                  seqLength,
                                                  args.useIndels,
                                                  args.use_reverse_complements)

        # Get the sequences for the given repertoire
        sequences = seqScrDict.keys()

        # Get the list of scores for the given repertoire
        scores = [seqScrDict[sequence] for sequence in sequences]

        # Create a network builder object
        netBuilder = NetworkBuilder(bitManip, args.use_reverse_complements)

        # Create the genotype network
        network = netBuilder.createGenoNet(repertoire, sequences, scores)

        # Get the number of components in the network
        numComponents = len(netBuilder.getComponents(network))

        # If there are more than one components,
        if numComponents > 1:
            # Get the giant component
            giant = netBuilder.getGiantComponent(network)

            # Set 'name' attribute for giant
            giant["name"] = repertoire + "_dominant"
        else:
            # The network and giant component are the same
            giant = network

        # Create a tuple in which to put the results
        netTuple = (repertoire, (network, giant))

        # Add the created network objects to the shared queue
        resultsQueue.put(netTuple)

        # Close the queue for this process
        resultsQueue.close()

    def _analyze_networks(self, repertoires, analyses):
        # Instantiate the analyzer. Initializing here makes it possible
        # to perform actions based on the list of requested analyses.
        self.analyzer = AnalysisHandler(self, analyses)

        # For each repertoire,
        for repertoire in repertoires:
            if self.VERBOSE:
                sys.stdout.write("\n" + repertoire + ":")
                sys.stdout.write("\n\t")

            # Perform the analysis
            self.analyzer.analyze(repertoire, analyses)

        if self.VERBOSE:
            print

    def _analyze_networks_parallel(self, repertoires, analyses):
        if self.VERBOSE:
            print

        # Make a copy of self
        self_copy = copy.deepcopy(self)

        # Delete the existing net dicts
        del self.repToNetDict
        del self.repToGiantDict

        # Re-initialize the deleted dicts
        self.repToNetDict = {}
        self.repToGiantDict = {}

        # Instantiate a concurrent queue for results
        resultsQueue = Queue()

        # Compute indices to be used in the loop
        indices = Genonets._process_blocks(len(repertoires), self.cmdArgs.num_procs)

        for i in indices:
            # Create separate processes for each repertoire in the current block
            processes = [
                Process(target=Genonets._analyze_gn,
                        args=(copy.deepcopy(self_copy),
                              analyses,
                              resultsQueue,
                              repertoires[j])
                        )
                for j in range(i - 1, Genonets._len_finished_reps(
                    i, len(repertoires), self.cmdArgs.num_procs))
            ]

            # Start the processes
            for p in processes:
                p.start()

            # Spin lock
            # FIXME: The condition in the loop can result in an infinite
            # iteration if one of the processes does not put results
            # in the queue. This condition should be replaced
            # with one that is reliable ...
            while len(self.repToNetDict) != Genonets._len_finished_reps(
                    i, len(repertoires), self.cmdArgs.num_procs):

                result = resultsQueue.get()

                if self.VERBOSE:
                    print("\tAnalysis results received for: " + result[0])

                self.repToNetDict[result[0]] = result[1][0]
                self.repToGiantDict[result[0]] = result[1][1]

    @staticmethod
    def _process_blocks(num_repertoires, num_processes):
        # If there are enough processes to process all repertoires
        # in parallel,
        if num_repertoires <= num_processes:
            # Only one iteration is required
            indices = [1]
        else:
            # Calculate the number of iteration that consume all
            # processes
            num_full_iters = num_repertoires / num_processes

            # Calculate the increment required in the last iteration
            incr_last = num_repertoires % num_processes

            # Create a list of increment values per iteration
            index_incrs = [num_processes for _ in range(num_full_iters)] \
                if num_full_iters != 0 else []

            # If required, add the last increment
            if incr_last != 0:
                index_incrs.append(incr_last)

            # Initialize the list of indices
            indices = [1]

            # Indices from index 1 onwards are calculated using prefix sum
            # of the increment list
            for i in range(0, len(index_incrs) - 1):
                indices.append(index_incrs[i] + indices[i])

        return indices

    @staticmethod
    def _len_finished_reps(cur_index, len_repertoires, max_procs):
        return min(len_repertoires, (cur_index - 1) + max_procs)

    def analyzeNets_parallel_old(self, repertoires, analyses):
        # Instantiate a concurrent queue for results
        resultsQueue = Queue()

        # Create separate processes for each repertoire
        processes = [
            Process(target=Genonets._analyze_gn,
                    args=(copy.deepcopy(self),
                          analyses,
                          resultsQueue,
                          repertoire)
                    )
            for repertoire in repertoires
        ]

        # Start the processes
        for p in processes:
            p.start()

        # Delete the existing net dicts
        del self.repToNetDict
        del self.repToGiantDict

        # Re-initialize the deleted dicts
        self.repToNetDict = {}
        self.repToGiantDict = {}

        # Spin lock
        # FIXME: The condition in the loop can result in an infinite
        #        iteration if one of the processes does not put results
        #        in the queue. This condition should be replaced
        #        with one that is reliable ...
        while len(self.repToNetDict) != len(repertoires):
            result = resultsQueue.get()

            print("Analysis results received for: " + result[0])

            self.repToNetDict[result[0]] = result[1][0]
            self.repToGiantDict[result[0]] = result[1][1]

    @staticmethod
    def _analyze_gn(genonetsCopy, analyses, resultsQueue, repertoire):
        # Initialize the AnalysisHandler object
        analyzer = AnalysisHandler(genonetsCopy, parallel=True)

        # Perform the analyses
        analyzer.analyze(repertoire, analyses)

        # Create a tuple in which to put the results
        resultTuple = (
            repertoire,
            (
                genonetsCopy.genotype_network(repertoire),
                genonetsCopy.dominant_network(repertoire)
            )
        )

        # Add results to the shared queue
        resultsQueue.put(resultTuple)

        # Close the queue for this process
        resultsQueue.close()

    # Description:	Returns the overlap matrix for all the genotype networks.
    # Return:		Overlap matrix as a list of lists.
    def _overlap_matrix(self):
        # Overlap matrix can only be computed if the networks have already
        # been created.
        if len(self.repToGiantDict) == 0:
            # Networks have not been created. Therefore, overlap matrix
            # cannot be computed.
            print("Overlap matrix cannot be computed before network creation.")

            return None

        # If the overlap matrix has already been computed,
        if self.analyzer.overlapMatrix:
            # No need to compute again, just return the existing matrix
            return self.analyzer.overlapMatrix
        else:
            # Perform the overlap computation
            self.analyzer.overlap()

            # Return the resulting matrix
            return self.analyzer.overlapMatrix




"""
    landscape_functions
    ~~~~~~~~~~~~~~~~~~~

    Wrapper for all landscape analysis functions.

    :author: Fahad Khalid
    :license: MIT, see LICENSE for more details.
"""

import igraph

from peak_functions import PeakAnalyzer
from path_functions import PathAnalyzer
from epistasis_functions import EpistasisAnalyzer
from genonets_utils import Utils


class Landscape:
    # Constructor
    def __init__(self, network, netUtils, seqToEscrDict, delta, bitManip):
        # Store reference to the network object
        self.network = network

        # Store reference to the netUtils object
        self.netUtils = netUtils

        # Create the peak analyzer
        self.peakAnalyzer = PeakAnalyzer(network, netUtils, delta)

        # Create the path analyzer
        self.pathAnalyzer = PathAnalyzer(network, netUtils, delta)

        # Create the epistasis analyzer
        self.epiAnalyzer = EpistasisAnalyzer(network, netUtils,
            seqToEscrDict, delta, bitManip)

        # Get a reference to the BitSeqManipulator in use
        self.bitManip = self.netUtils.bitManip

    # ----------------------------------------------------------------
    # Peak analysis methods
    # ----------------------------------------------------------------

    def getPeaks(self, recompute):
        return self.peakAnalyzer.get_all_peaks(recompute)

    # ----------------------------------------------------------------
    # Path analysis methods
    # ----------------------------------------------------------------

    def getAccessiblePaths(self, pathLength=0):
        return self.pathAnalyzer.getAccessiblePaths(pathLength)

    # ----------------------------------------------------------------
    # Epistasis analysis methods
    # ----------------------------------------------------------------

    def getEpistasis(self):
        return self.epiAnalyzer.getEpiAll()

    # ----------------------------------------------------------------
    # Calculate mutational distances from summit for all vertices
    # ----------------------------------------------------------------

    def populateDistsToSummit(self):
        # Get the summit sequence
        summit = Utils.getSeqWithMaxScore(
            self.network, self.bitManip.seqLength)

        # Get vertex that represents summit
        trgtVrtx = self.netUtils.getVertex(summit, self.network)

        self.network.vs["Distance from Summit"] = [
            len(
                self.network.get_shortest_paths(
                    srcVrtx, to=trgtVrtx, weights=None,
                    mode=igraph.OUT, output="epath")[0]
            )
            for srcVrtx in self.network.vs
        ]



"""
    This module implements the functions required for identification of peaks
    in a network (when viewed as a fitness landscape).

    Note: The term 'peak' is used in this module to refer to both peaks and
          plateaus.

    The algorithm:
        1. Sort all genotypes in descending order of score.
        2. For each genotype,
        3.      Create a list of all genotypes that are in the same score band
                as the focal genotype. Exclude genotypes that have already been
                processed.
        4.      Create the neutral zone, i.e., a list of all genotypes that are
                in the score band and connected. The focal genotype is also
                included in the neutral zone.
        5.      If any genotype in the neutral zone has a 1-neighbor outside
                the neutral zone that has already been processed (either as a
                peak member of as a valley member), the entire neutral zone is
                a non-peak.
        6.      Else, create a new peak such that all genotypes in the neutral
                zone are members of the peak.
        7.      Mark all genotypes in the neutral zone as processed.

"""

from genonets_utils import Utils


class PeakAnalyzer:
    """
    Encapsulates the computations required for identifying peaks in the given
    network. It is assumed that the given network is a single connected
    component; typically the dominant component of a genotype network.

    """

    def __init__(self, network, net_utils, delta):
        """
        Initializes the object.

        :param network: (igraph.Graph) The network within which to identify
                peaks. The network must be connected, i.e., it must not consist
                of more than one connected component.
        :param net_utils: (genonets.graph_utils.NetworkBuilder) Reference to
                the object providing the required network manipulation
                functions.
        :param delta: (float) The experimental noise value to consider when
                evaluating each genotype for optimality.

        """

        self._delta = delta
        self._network = network
        self._net_utils = net_utils

        # Reference to the BitSeqManipulator in use
        self._bit_manip = net_utils.bitManip

        # Map { genotype: {all 1-neighbors} }
        self._neighbor_map = {
            s: set(self._net_utils.getNeighborSequences(s, self._network))
            for s in network.vs['sequences']
        }

        # Set of elements that have already been processed, either individually
        # or as part of a neutral zone for another genotype
        self._processed_genotypes = set()

        # Map { peak_id: {genotypes} }. To be populated with all computed peaks.
        self._peaks = None

    def get_peak_with_genotype(self, genotype, recompute=False):
        """
        Returns the peak object of which the given genotype is a member.

        :param genotype: (str) The genotype to search in the peaks.
        :param recompute: (bool) 'True' if peaks should be computed again even
                if the corresponding analyzer has been used to compute peaks
                already. 'False' otherwise.

        :return: (dict) of the form {peak_id: {all genotypes in the peak}} for
                the peak in which the given genotype was found.
                (None) if the genotype could not be found in any peak. This
                would be the case if the genotype lies in a valley.

        """

        # Get all the peaks
        peaks = self.get_all_peaks(recompute)

        for peak_id in peaks:
            # Get the set of genotypes that belong to this peak
            peak_members = peaks[peak_id]

            # If the given genotype is a member,
            if genotype in peak_members:
                return {peak_id: peak_members}

        return None

    def get_all_peaks(self, recompute=False):
        """
        Returns all peaks in the network with which the corresponding analyzer
        object was initialized.

        If the corresponding analyzer object has been used to compute peaks
        already, and 'recompute' is 'False', the already computed peaks object
        is returned. Otherwise, peaks are computed first and then returned.

        :param recompute: (bool) 'True' if peaks should be computed again even
                if the corresponding analyzer has been used to compute peaks
                already. 'False' otherwise.

        :return: (dict) of the form {peak_id: {all genotypes in the peak}}.

        """

        if not self._peaks or recompute:
            # Structured array of tuples (genotype, score), sorted in
            # descending order of scores
            genotype_score_pairs = Utils.getSortedSeqEscArr(
                self._network,
                self._bit_manip.seqLength,
                sortOrder='descending'
            )

            # Identify peaks
            self._peaks = self._build_peaks(genotype_score_pairs)

        return self._peaks

    def _build_peaks(self, elements):
        """
        Assuming the given array of (genotype, score) tuples is already sorted
        in descending order of scores, this function orchestrates the algorithm
        for determining whether a given genotype lies in a peak or a valley.
        That is, this function implements steps 2 - 7 of the overall algorithm
        specified in the documentation of this module.

        :param elements: (list of tuples), where each tuple is of the form
                    (genotype, score). The list is assumed to be in descending
                    order of scores.

        :return: (dict) { key=peak_id: value={all genotypes in the peak} }

        """

        peaks = {}

        for i in xrange(len(elements)):
            # If the current element has already been processed,
            if elements[i]['sequence'] in self._processed_genotypes:
                # Move on to the next element
                continue

            # List of elements for which the score lies within the score band
            # defined by the score of the current focal element
            score_band = self._build_score_band(i, elements)

            # Neutral zone, i.e., all elements within the score band that are
            # connected
            neutral_zone = self._build_neutral_zone(elements[i], score_band)

            # Append the focal element itself to the neutral zone.
            # Note: Insert at index 0 would be nicer, but the cost would O(n).
            neutral_zone.append(elements[i])

            # If any genotype in the neutral zone has a 1-neighbor outside
            # the neutral zone that has already been processed (either as a
            # peak member of as a valley member),
            if self._neighbors_processed_element(neutral_zone):
                # The entire neutral zone is a non-peak. There's nothing more
                # to be done.
                pass
            else:
                # Create a new peak, and add all members of the neutral zone
                # to this peak
                peak_id = len(peaks)
                peaks[peak_id] = self._create_peak(neutral_zone)

            # Mark all elements in the neutral zone as processed, i.e., there
            # is no need to process each one individually
            self._processed_genotypes.update({
                e['sequence'] for e in neutral_zone
            })

        return peaks

    def _build_score_band(self, index_of_focal, elements):
        """
        Returns a list of elements that fall within the noise determined score 
        band for the given focal element (excluding elements that have already
        been processed), where the focal element is identified by the given
        index.

        :param index_of_focal: (int) Index of the focal element in the given
                list of all elements.
        :param elements: (numpy.array) Array of (genotype, score) tuples
                corresponding to all genotypes in the network in consideration.
        
        :return: (list) Elements for which the score lies within the band of
                the focal element, and which have not been processed already.
        
        """

        score_band = []
        focal_element = elements[index_of_focal]

        # Starting from the element succeeding the focal element,
        for i in xrange(index_of_focal + 1, len(elements)):
            # If the score lies within the band,
            if elements[i]['escore'] >= focal_element['escore'] - self._delta:
                # If the current element has already been processed, it means
                # that it is either part of an existing peak, or an existing
                # non-peak. It can only be indirectly connected to the current
                # score band via an existing peak, or via an existing non-peak.
                # Either way, we can safely ignore it, as other checks should
                # be sufficient to accurately categorize the current score band.
                if elements[i]['sequence'] in self._processed_genotypes:
                    pass
                else:
                    score_band.append(elements[i])
            else:
                # Once we've reached an element which lies in a lower band, we
                # know that the remaining elements will also lie in lower
                # bands. Therefore, we don't need to look any further.
                break

        return score_band

    def _build_neutral_zone(self, focal_element, score_band):
        """
        Returns the neutral zone for the given focal element.

        The neutral zone consists of all elements that satisfy the following
        conditions:
            1) The element lies within the score band for the given focal 
               element.
            2) The element is either a 1-neighbor of the focal element, or it
               is indirectly connected to the focal element, such that all
               elements that constitute the path between this element and the
               focal element lie within the score band of the focal element.
        
        :param focal_element: (tuple) of form (genotype, score) corresponding
                to the element for which the score band was constructed.
        :param score_band: (list) of (genotype, score) tuples, where each
                tuple represents a genotype for which the score lies in the
                score band for the focal element.
        
        :return: (list) The neutral zone.
        
        """

        # 1-neighbors of the focal element
        neighbors_of_focal = self._get_neighbors_of(focal_element['sequence'])

        # Elements in the given score band that are also in the neighborhood of
        # the focal element. These are by definition already part of the neutral
        # zone.
        neutral_zone = [
            item for item in score_band
            if item['sequence'] in neighbors_of_focal
        ]

        # Elements in the given score band that are not 1-neighbors of the
        # focal element
        band_non_neighbors = [e for e in score_band if e not in neutral_zone]

        # Perform a breadth-first search of the neutral zone to find those
        # elements in band-non-neighbors that might be indirectly connected
        # (via other members of the neutral zone) to the focal element.
        neutral_zone = self._bfs(neutral_zone, band_non_neighbors)

        return neutral_zone

    def _bfs(self, neutral_zone, band_non_neighbors):
        """
        Returns the complete neutral zone by performing a breadth-first search
        of the neutral zone to check if any of the elements in
        score-band-non-neighbors are connected to any of the neutral zone
        members. The search is dynamic, i.e., as connected elements are found,
        these are themselves then considered members of the neutral zone, and
        used as seed elements for further search.

        :param neutral_zone: (list) of tuples of the form (genotype, score),
                which represents the neutral zone to use as a starting point for
                the BFS.
        :param band_non_neighbors: (list) of tuples of the form
                (genotype, score). This is the list of elements against which to
                check connectivity of elements of the given neutral zone.

        :return: (list) The extended neutral zone.

        """

        i = 0  # Index variable

        # Keep iterating for as long as there are elements left in the
        # neutral zone
        while i < len(neutral_zone):
            # Set of members of score-band-non-neighbors that are found to be
            # indirect neighbors of the focal element, and should therefore be
            # removed from the list of score-band-non-neighbors
            non_neighbors_to_remove = []

            # For each member of score-band-non-neighbors,
            for non_neighbor in band_non_neighbors:
                # Flag to indicate whether the non-neighbor is connected to the
                # ith element in the neutral zone,
                are_connected = self._are_neighbors(
                    neutral_zone[i]['sequence'],
                    non_neighbor['sequence']
                )

                # If the two are connected,
                if are_connected:
                    # Append the non-neighbor to the neutral zone
                    neutral_zone.append(non_neighbor)

                    # Mark the non-neighbor, so that it is removed from the
                    # set of score-band-non-neighbors after the for loop.
                    non_neighbors_to_remove.append(non_neighbor)

            # Remove those elements from the set of score-band-non-neighbors
            # that have already been added to the neutral zone
            band_non_neighbors = [
                e for e in band_non_neighbors
                if e not in non_neighbors_to_remove
            ]

            # Move on to the next element in the neutral zone
            i += 1

        return neutral_zone

    def _neighbors_processed_element(self, neutral_zone):
        """
        Checks if any member of the given neutral zone has a 1-neighbor outside
        the neutral zone that has already been processed (either as a member of
        a peak, or a member of a valley).

        :param neutral_zone: (list) of tuples of the form (genotype, score).

        :return: (bool) 'True' if any member of the neutral zone has an already
                    processed element as 1-neighbor . 'False' otherwise.

        """

        for element in neutral_zone:
            # All 1-neighbors of the element
            neighbors = self._get_neighbors_of(element['sequence'])

            # If any neighbor has already been processed,
            if not neighbors.isdisjoint(self._processed_genotypes):
                return True

        return False

    @staticmethod
    def _create_peak(neutral_zone):
        """
        Extracts all genotypes from the given neutral zone and returns them as
        a single collection that represents all members of a new peak.

        :param neutral_zone: (list) of tuples of the form (genotype, score).

        :return: (set) All genotypes in the peak.

        """

        return {e['sequence'] for e in neutral_zone}

    def _get_neighbors_of(self, genotype):
        """
        Returns all 1-neighbors of the given genotype.

        :param genotype: (str) Genotype for which to return the neighbors.

        :return: (set) All 1-neighbors of the given genotype.

        """

        return self._neighbor_map[genotype]

    def _are_neighbors(self, g_1, g_2):
        """
        Checks if g_2 is a 1-neighbor of g_1.

        :param g_1: (str) The first genotype.
        :param g_2: (str) The second genotype.

        :return: (bool) 'True' if g_2 is a 1-neighbor of g_1. 'False' otherwise.

        """

        return True if g_2 in self._get_neighbors_of(g_1) else False



"""
    overlap_functions
    ~~~~~~~~~~~~~~~~~

    Contains functions used for computation of overlap.

    :author: Fahad Khalid
    :license: MIT, see LICENSE for more details.
"""


class OverlapAnalyzer:
    # Constructor
    def __init__(self, repToGiantDict, repertoires, bitManip, isDoubleStranded, ordering_function):
        # Copy of dict: key=repertoire, value=giant
        self.repToGiantDict = repToGiantDict

        # List of available repertoires
        self.repertoires = repertoires

        # Reference to the bit manipulator object
        self.bitManip = bitManip

        # Flag to indicate whether reverse complements should be
        # considered for genotypes.
        self.isDoubelStranded = isDoubleStranded

        # Sort the list of genotype sets such that the order in the list
        # is the same as that in the input input file
        self.repertoires.sort(key=ordering_function)

    def getOverlapData(self):
        if len(self.repertoires) < 2:
            print("Overlap computation triggered with only one repertoire!")
            print("Overlap can only be calculated with 2 or more repertoires.")

            return None, None

        # Overlap dict for all repertoires. Dict{rep : {seq : [target reps]}}
        allOverlap = {
            rep: {
                seq: []
                for seq in self.repToGiantDict[rep].vs["sequences"]
            }
            for rep in self.repertoires
        }

        # Initialize the overlap matrix with zeros
        overlapMat = [
            [0 for _ in range(len(self.repertoires))]
            for _ in range(len(self.repertoires))
        ]

        # For each repertoire,
        for i in range(len(self.repertoires) - 1):
            # Get giant for this repertoire
            giant_i = self.repToGiantDict[self.repertoires[i]]

            # Get sequence list from giant
            seqs_i = giant_i.vs["sequences"]

            # For the rest of the repertoires,
            for j in range(i + 1, len(self.repertoires)):
                # Get giant for this repertoire
                giant_j = self.repToGiantDict[self.repertoires[j]]

                # Get sequence list from giant
                seqs_j = giant_j.vs["sequences"]

                # Get the list of sequences that are common to both repertoires
                overlapList = self.getOverlapList(seqs_i, seqs_j)

                # Get the number of sequences that overlap between the
                # two repertoires, and set it in the matrix
                overlapMat[i][j] = overlapMat[j][i] = len(overlapList)

                # If there is at least one shared sequence,
                if len(overlapList) > 0:
                    # For each sequence in the list,
                    for sequence in overlapList:
                        # Append the target to the target list for this sequence
                        allOverlap[self.repertoires[i]][sequence].append(self.repertoires[j])
                        allOverlap[self.repertoires[j]][sequence].append(self.repertoires[i])

        return overlapMat, self.repertoires, allOverlap

    # Returns a list of sequences that are common to the two input sequence
    # lists
    def getOverlapList(self, seqs1, seqs2):
        # If reverse complements should be considered,
        if self.isDoubelStranded:
            # Reverse complements of 'seqs2' in bit format
            rc_seqs2 = [
                self.bitManip.getReverseComplement(self.bitManip.seqToBits(s))
                for s in seqs2
            ]

            # Reverse complements in string format
            rc_seqs2 = [self.bitManip.bitsToSeq(rc) for rc in rc_seqs2]

            # Add the reverse complements to 'seqs2', while ensuring uniqueness
            # of each added sequence.
            seqs2.extend(list(set(seqs2) | set(rc_seqs2)))

        return list(set(seqs1) & set(seqs2))



"""
    evolvability_functions
    ~~~~~~~~~~~~~~~~~~~~~~

    Contains functions used for genotype and phenotype evolvability computations.

    :author: Fahad Khalid
    :license: MIT, see LICENSE for more details.
"""


class EvolvabilityAnalyzer:
    # Constructor
    def __init__(self, network, dataDict, seqToRepDict, repToGiantDict,
                 rcToSeqDict, bitsToSeqDict, netBuilder,
                 isDoubleStranded):
        # Reference to the network on which to perform this
        # analysis
        self.network = network

        # Reference to dict: key=rep, value=dict{seq: score}
        self.dataDict = dataDict

        # Copy dict: key=sequence, value=[repertoires]
        self.seqToRepDict = seqToRepDict

        # Copy of dict: key=repertoire, value=giant
        self.repToGiantDict = repToGiantDict

        # Dict {bit reverse complement: string seq} - Holds reverse
        # complement (bit format) to original sequence (string)
        self.rcToSeqDict = rcToSeqDict

        # Dict {bitseq : seq}
        self.bitsToSeqDict = bitsToSeqDict

        # Get a reference to the NetworkUtils object
        self.netBuilder = netBuilder

        # Flag indicating whether the genotypes correspond to double
        # stranded molecules
        self.isDoubleStranded = isDoubleStranded

        # Reference to the bit manipulator object
        self.bm = self.netBuilder.bitManip

        # Reference to the list of evolvability values for all sequences
        self.evoTuples = None

    @staticmethod
    def updateSeqToRepDict(seqToRepDict_original, repToGiantDict):
        # Make a deep copy of the dict
        seqToRepDict = seqToRepDict_original.copy()

        # For each sequence key in the dict,
        for seq in seqToRepDict.keys():
            # Update the list of repertoires corresponding to this
            # sequence with only those repertoires for which this
            # sequence exists in the giant.
            seqToRepDict[seq] = [
                rep for rep in seqToRepDict[seq]
                if seq in repToGiantDict[rep].vs["sequences"]
            ]

        return seqToRepDict

    @staticmethod
    def buildRcToSeqDict(seqToRepDict, bm):
        # Initialize the dict
        rcToSeqDict = dict()

        # For each sequence key in the dict,
        for seq in seqToRepDict.keys():
            # Compute the reverse complement
            rcBitSeq = bm.getReverseComplement(
                bm.seqToBits(seq))

            # With the reverse complement bit sequence as the key,
            # add the original sequence (string format) as the value.
            rcToSeqDict[rcBitSeq] = seq

        return rcToSeqDict

    @staticmethod
    def buildBitsToSeqDict(seqToRepDict, rcToSeqDict, bm, isDoubleStranded):
        # Create the dictionary - {bit sequence: sequence}
        bitsToSeqDict = {
            bm.seqToBits(seq): seq for seq in seqToRepDict.keys()
        }

        # If we need to consider reverse complements,
        if isDoubleStranded:
            # Use the 'reverse complement to sequence dict' to get the
            # the reverse complements corresponding to all the
            # sequences.
            for rc in rcToSeqDict.keys():
                bitsToSeqDict[rc] = bm.bitsToSeq(rc)

        return bitsToSeqDict

    # Returns repertoire evolvability
    def getReportoireEvo(self):
        # Get evo tuples for all sequences in the network
        evoTuples = self.getEvoAll()

        # Get repertoire lists from the tuples
        repLists = [
            evoTuples[1].keys()
            for evoTuples in evoTuples if evoTuples[1]
        ]

        # Combine all repLists into one list
        targets = []
        for repList in repLists:
            targets.extend([target for target in repList])

        # Remove redundant repertoires
        targets = list(set(targets))

        try:
            evolvability = float(len(targets)) / float(len(self.dataDict) - 1)

            return evolvability, targets
        except ZeroDivisionError:
            return 0, targets

    def getEvoAll(self, recompute=False):
        # If either the evolvability values have not been computed already,
        # or the caller has explicitly requested re-computing,
        if not self.evoTuples or recompute:
            # Get tuples: (evo, targetRepsDict) for all sequences in the network
            self.evoTuples = [
                self.getSeqEvo(seq)
                for seq in self.network.vs["sequences"]
            ]

        return self.evoTuples

    # Returns evolvability score for the given sequence
    def getSeqEvo(self, sequence):
        # Get a list of sequences that are 1-neighbors, but not part of the
        # genotype network
        externNeighbors = self.netBuilder.getExternalNeighbors(sequence, self.network)

        # Get dictionary: key=repertoire, value=extNeighs. The keys represents the
        # list of repertoires to which the given sequence could join by undergoing
        # a single mutation.
        targetReps = self.getEvoTargetReps(externNeighbors)

        # Compute and return the evolvability score, i.e., No. of repertoire
        # targets found for this sequence / No. of repertoires available in
        # the dataset.
        # FIXME: ZeroDivisionError can only occur if the denominator is zero.
        # This already be checked at the very beginning of the function to
        # avoid the rest of the computation ...
        try:
            evolvability = float(len(targetReps)) / float(len(self.dataDict) - 1)
        except ZeroDivisionError:
            evolvability = 0

        return (evolvability, targetReps)

    # For the given sequence, list of external neighbors, and the sequence
    # to repertoire dictionary, return a list of repertoires for which the given
    # sequence could develop score >= tau by a single mutation.
    def getEvoTargetReps(self, extNeighs):
        targetReps = dict()

        # For each external neighbor,
        for extNeigh in extNeighs:
            try:
                # Get the sequence corresponding to the bitseq
                extNeighSeq = self.bitsToSeqDict[extNeigh]
            except KeyError:
                # If the bitseq corresponding to sequence could not be found,
                # it means that it does not exist in any other repertoire. We
                # can therefore safely skip the current iteration
                continue

            # If the external neighbor is in any repertoire,
            if extNeighSeq in self.seqToRepDict:
                # Append the corresponding repertoire(s) to the
                # list of target repertoires for this sequence.
                self.appendToTargets(extNeighSeq, targetReps)
            elif self.isDoubleStranded:
                # Get the bit representation for the extNeighSeq
                extNeighBits = self.bm.seqToBits(extNeighSeq)

                # If the external neighbor is amongst the reverse complements,
                if extNeighBits in self.rcToSeqDict:
                    # Get the string representation of the reverse complement
                    # of the external neighbor sequence.
                    # Note: The reason we are using the original sequence rather
                    # than the reverse complement is that if we use the
                    # reverse complement, the visualization in the Genonets
                    # Server would not be able to highlight the target node.
                    strSeq = self.rcToSeqDict[extNeighBits]

                    self.appendToTargets(strSeq, targetReps)

        return targetReps

    def appendToTargets(self, seq, targetReps):
        # For each repertoire that contains the external neighbor,
        for repertoire in self.seqToRepDict[seq]:
            # The focal repertoire should not be considered
            if self.network["name"] in [repertoire, repertoire + "_dominant"]:
                continue

            # We have found a repertoire with a genotype network that
            # has a 1-neighbor to which  the sequence can evolve.
            if repertoire not in targetReps:
                targetReps[repertoire] = []

            targetReps[repertoire].append(seq)



#!/usr/bin/env python

from setuptools import setup

# Read the contents of the README file
import os
this_directory = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(this_directory, 'README.md')) as f:
    long_description = f.read()

setup(
    name='genonets',
    version='1.1.10',
    description='Framework for creating and analyzing genotype networks from '
                'data.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Fahad Khalid',
    author_email='fahad.khalid@ieu.uzh.ch',
    url='https://github.com/fkhalid/genonets',
    license='MIT',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2.7',
        'Operating System :: OS Independent',
        'Natural Language :: English',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
    ],
    keywords='genonets genotype phenotype network',
    packages=['genonets'],
    package_data={
        'genonets': [
            'sample/*.py',
            'sample/data/*.txt'
        ]
    },
    install_requires=[
        'python-igraph==0.7.1.post6',
        'numpy>=1.8.2'
    ]
)
